# Copyright: 2007-2013 Paul Obermeier (obermeier@poSoft.de)
# Distributed under BSD license.

set docDir   [pwd]
set cawtDir  [file join $docDir ".."]
set finalDir [file join $docDir "Final"]

if { ! [file isdirectory $finalDir] } {
    file mkdir $finalDir
}

set auto_path [linsert $auto_path 0 $cawtDir [file join $cawtDir "Externals"]]

# Possible targets: all, ref, user
set genDoc "all"
if { $argc > 0 } {
    set genDoc [lindex $argv 0]
}

set retVal [catch {package require cawt} pkgVersion]

puts [format "%-25s: %s" "Tcl version" [info patchlevel]]
puts [format "%-25s: %s" "Cawt version" $pkgVersion]
puts [format "%-25s: %s" "Twapi version" [::Cawt::GetPkgVersion "twapi"]]

if { $genDoc eq "ref" || $genDoc eq "all" } {
    cd $cawtDir

    package require ruff

    puts "Generating reference documentation from source code ..."
    ::ruff::document_namespaces html \
        [list ::Cawt ::Word ::Excel ::Explorer ::Ppt ::Ocr ::Matlab ::Earth] \
        -title "CAWT Reference" \
        -output [file join $finalDir "CawtReference-$pkgVersion.html"]
    cd $docDir
}

if { $genDoc eq "user" || $genDoc eq "all" } { 
    cd $docDir

    puts "Generating user manual from Word and PowerPoint files ..."
    set pptInFile  [file join [pwd] "UserManual" "CawtFigures.ppt"]
    set wordInFile [file join [pwd] "UserManual" "CawtManualTemplate.doc"]

    set userManFile [file join $finalDir "CawtManual-$pkgVersion.doc"]

    set outFigureDir [file join $finalDir "CawtFigures"]
    set testDir [file join $cawtDir "TestPrograms"]

    # Generate the figures for the user manual from the PowerPoint file.
    ::Ppt::ExportPptFile $pptInFile $outFigureDir "Figure-%02d.png" 1 end "PNG" -1 -1 false false

    # Open the user manual template and perform the following actions:
    #   Insert a table with the test programs available (from folder TestPrograms).
    #   Insert the generated figures replacing the placeholder text.
    # Then save the finished user manual in the Final folder.
    set wordId [::Word::OpenNew]
    set docId [::Word::OpenDocument $wordId $wordInFile true]
    ::Word::SetCompatibilityMode $docId $::Word::wdWord2003

    set placeHolder "%TABLE TestPrograms%"
    set numTables [::Word::GetNumTables $docId]
    set foundPlaceHolder false
    for { set n 1 } { $n <= $numTables } {incr n } {
        set tableId [::Word::GetTableIdByIndex $docId $n]
        # Placeholder must be listed in row 2, column 1.
        set cellCont [::Word::GetCellValue $tableId 2 1]
        if { $cellCont eq $placeHolder } {
            puts "    Replacing placeholder \"$placeHolder\" with list of test programs."
            set testFileList [lsort [glob -directory $testDir Earth* Excel* Explorer* Matlab* Ocr* Ppt* Word*]]
            set numRows [::Word::GetNumRows $tableId]
            set missingRows [expr [llength $testFileList] - $numRows +1]
            for { set r 1 } { $r <= $missingRows } {incr r } {
                ::Word::AddRow $tableId
            }

            set row 2
            foreach testFile $testFileList {
                set fp [open $testFile "r"]
                set description ""
                while { [gets $fp line] >= 0 } {
                    if { $line eq "#" } break
                    append description [string trim [string trim $line "#"]]
                }
                close $fp
                set f [file tail $testFile]
                ::Word::SetCellValue $tableId $row 1 $f
                ::Word::SetCellValue $tableId $row 2 $description
                incr row
            }
            set foundPlaceHolder true
            break
        }
    }
    if { ! $foundPlaceHolder } {
        puts "    Warning: Placeholder \"$placeHolder\" not referenced in Word-Document"
    }

    foreach fig [glob -directory $outFigureDir *] {
        set figure [file tail $fig]
        set placeHolder "%FIGURE $figure%"
        set myRange [::Word::GetStartRange $docId]
        if { ! [::Word::FindString $myRange $placeHolder] } {
            puts "    Warning: Figure $figure not referenced in Word-Document"
            continue
        } else {
            puts "    Replacing keyword $placeHolder with figure $figure"
        }
        set rangeId [::Word::GetSelectionRange $docId]
        ::Word::InsertImage $rangeId $fig
        # TODO Selection.InlineShapes(1).PictureFormat.CropBottom = 234.26
    }
    ::Word::SaveAs $docId $userManFile
    ::Word::Quit $wordId

    ::Cawt::Destroy
}

exit 0
